/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 932 $ $LastChangedDate: 2002-09-11 17:06:10 +0200 (Wed, 11 Sep 2002) $
 *
 * Addons library for Gont, Xml module C implementation.
 */

#include <gont/runtime.h>

#include <libxml/parser.h>

struct gont_string *Xml___xmlDocDumpMemory(xmlDocPtr doc, gont_int format)
{
	int size = 0;
	xmlChar *ptr = NULL;
	struct gont_string *ret;
	
	xmlIndentTreeOutput = 1;
	xmlDocDumpFormatMemory(doc, &ptr, &size, format ? 1 : 0);

	if (ptr == NULL)
		gont_invalid_argument("Xml___xmlDocDumpMemory");
	
	ret = gont_string_of_utf8(ptr, size);

	xmlFree(ptr);

	return ret;
	
}


struct node;

struct node *Xml$$make_text_node(struct gont_string*);
struct node *Xml$$make_element_node(struct gont_string*, struct gont_union *,
	struct gont_union *);
struct gont_union *List$$nil(void);
struct gont_union *List$$cons(gont_int, struct gont_union *);

static int is_text_node(xmlNodePtr n)
{
	switch (n->type) {
	case XML_TEXT_NODE:
	case XML_CDATA_SECTION_NODE:
	case XML_ENTITY_REF_NODE:
		return 1;
	default:
		return 0;
	}
}

static struct gont_union *get_props(xmlDocPtr doc, xmlNodePtr n)
{
	xmlAttrPtr a;
	struct gont_union *attrs = List$$nil();

	for (a = n->properties; a && a->next; a = a->next)
		/* nothing */ ;
		
	while (a) {
		struct gont_string *name, *value;
		xmlChar *utf_val;
		name = gont_string_utf8(a->name);
		utf_val = xmlNodeListGetString(doc, a->children, 1);
		value = gont_string_utf8(utf_val);
		xmlFree(utf_val);
		attrs = List$$cons((gont_int)gont_make_tuple(2, name, value), 
				   attrs);
		a = a->prev;
	}

	return attrs;
}

static struct node *convert_node(xmlDocPtr doc, xmlNodePtr n, int skip_spaces)
{
	xmlNodePtr c, set_in, set_to;
	struct gont_union *children;
	struct node *cc;

	if (is_text_node(n)) {
		xmlChar *content;

		if (skip_spaces) {
			xmlNode *p;
			for (p = n; p; p = p->next)
				if (!xmlIsBlankNode(p))
					break;
			if (p == NULL)
				return NULL;
		}
		
		content = xmlNodeListGetString(doc, n, 1);
		cc = Xml$$make_text_node(gont_string_utf8(content));
		xmlFree(content);
		return cc;
	}
	
	children = List$$nil();
	
	for (c = n->children; c && c->next; c = c->next)
		/* nothing */ ;
	
	while (c) {
		if (is_text_node(c)) {
			/* Save next. */
			set_in = c;
			set_to = c->next;
			
			/* Mark for xmlNodeListGetString(). */
			c->next = NULL;
			
			if (c->prev == NULL) {
				cc = convert_node(doc, c, skip_spaces);
				c = NULL;
			} else {
				while (c->prev && is_text_node(c))
					c = c->prev;
				cc = convert_node(doc, c->next, skip_spaces);
			}
			
			/* Restore. */
			set_in->next = set_to;
		} else {
			cc = convert_node(doc, c, skip_spaces);
			c = c->prev;
		}

		/* Prepend to list. */
		if (cc)
			children = List$$cons((gont_int)cc, children);
	}

	cc = Xml$$make_element_node(gont_string_utf8(n->name), children, 
				  get_props(doc, n));

	return cc;
}

struct node *Xml___convert_document(xmlDocPtr doc, gont_int skip_spaces)
{
	xmlNodePtr n;
	struct node *ret;

	if (doc == NULL)
		gont_invalid_argument("Xml___convert_document");

	for (n = doc->children; n; n = n->next)
		if (n->type == XML_ELEMENT_NODE)
			break;
	
	if (n == NULL)
		gont_invalid_argument("Xml___convert_document");
	
	
	ret = convert_node(doc, n, skip_spaces ? 1 : 0);

	return ret;
}

void Xml___set_root(xmlDocPtr doc, xmlNodePtr node)
{
	doc->children = node;
}

void Xml___libxml_bug_workaround(xmlNodePtr node)
{
	node->type = XML_TEXT_NODE;
}
