/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 840 $ $LastChangedDate: 2002-09-01 16:56:25 +0200 (Sun, 01 Sep 2002) $
 *
 * Cg_util - utilities for code generation.
 */

open Ast;
open Util;

Stdio::FILE ksi_sink = Stdio::stdout;

Ksi::ty ptroff = Ksi::Macro_type(make_id("ptr_off"));
Ksi::ty ksi_void = Ksi::Void;
Ksi::ty c_double = Ksi::Macro_type(make_id("c_double"));
Ksi::ty c_double_ptr = Ksi::Ptr(c_double);
Ksi::ty gont_string = Ksi::Ptr(Ksi::Struct(make_id("gont_string")));
Ksi::ty gont_lambda_expr = Ksi::Ptr(Ksi::Struct(make_id("gont_lambda_expr")));
Ksi::ty gont_union = Ksi::Ptr(Ksi::Struct(make_id("gont_union")));
Ksi::ty tuple_ptr = Ksi::Ptr(Ksi::Poly);
Ksi::ty poly = Ksi::Poly;
Ksi::id our_closure_id = make_id("closure$");
Ksi::id parent_closure_id = make_id("parent_closure$");
Ksi::id dummy_id = make_id("__dummy$__");
Ksi::ty null_type = Ksi::Ptr(Ksi::Void);
Ksi::expr null_value = Ksi::Null(null_type);
Ksi::expr nop = Ksi::Nop;
Ksi::expr true_const = Ksi::Bool_const(true);
Ksi::expr false_const = Ksi::Bool_const(false);
Ksi::expr zero_const = Ksi::Int_const(ptroff, Int64::zero);
Ksi::expr one_const = Ksi::Int_const(ptroff, Int64::one);

void output_decl(Ksi::decl d)
{
    Ksi::output_decl(ksi_sink, d);
}

void top_comment(string s)
{
    output_decl(Ksi::Top_comment(s));
}

Ksi::id make_id(string s)
{
    return Ksi::make_id(s);
}

Ksi::expr call(string name, <Ksi::expr>list args)
{
    return Ksi::Call(make_id(name), args);
}

Ksi::expr gont_alloc(Ksi::expr sz)
{
    return call("gont_xmalloc", [sz]);
}

Ksi::expr bin(Ksi::binop o, Ksi::expr e1, Ksi::expr e2)
{
    return Ksi::Binary(o, e1, e2);
}

Ksi::expr pair(Ksi::expr e1, Ksi::expr e2)
{
    return Ksi::Compound(List::cons2(e1, e2));
}

Ksi::expr make_vardef(Ksi::ty t, Ksi::id n, Ksi::expr ini)
{
    return pair(Ksi::Var(t, n), Ksi::Modify(Ksi::Ref(n), ini));
}

<Ksi::attr>list public_attrs(bool is_public)
{
    if (is_public)
	return [Ksi::Public];
    else
	return [Ksi::Local];
}

Ksi::expr raddr(Ksi::expr e)
{
    return Ksi::Unary(Ksi::R_addr, e);
}

Ksi::expr wfl_locate('a c, *(Ksi::expr ('a, 'b)) f, <'b>located v)
{
    push_location(v.loc);
    try {
	let r = f(c, v.a) in
	return Ksi::Line(v.loc.first_line, r);
    } finally {
	pop_location();
    }
}

Ksi::expr symbol_union_tag(var_symbol s)
{
    let id = make_id(s.a.extern_name) in
    if (has_property(s, Is_exception))
	return Ksi::Ref(id);
    else
	return Ksi::Enum_ref(id);
}

Ksi::expr tuple_nth(Ksi::expr e, int n)
{
    let pos = bin(Ksi::Mult, 
		  Ksi::Int_const(ptroff, Int64::of_int(n)),
		  Ksi::Convert(ptroff, Ksi::Size_of(ptroff))) in
    return Ksi::Indirect(bin(Ksi::Plus, Ksi::Convert(tuple_ptr, e), pos));
}

Ksi::expr union_tag(Ksi::expr e)
{
    return Ksi::Arrow(Ksi::Convert(gont_union, e), make_id("tag"));
}

Ksi::expr union_value(Ksi::expr e)
{
    return Ksi::Arrow(Ksi::Convert(gont_union, e), make_id("value"));
}

// vim:sts=4:sw=4
