/**
+ RPN calculator

We are now going to write Reverse Polish Notation calculator.
RPN calculator contains stack of numbers, it reads its input word
by word.  When it reads number it pushes it on the stack.  When it reads
operator like "+" or "-", it pops two numbers from stack, performs operation
on them and pushes result back on the stack.  There are few special
operators, like "p" that prints top of stack or "c" that clears the stack.
In Unix dc(1) program is arbitrary precision RPN calculator.
*/

// Here we define stack of integers.  Skip the bizzare type for now.
<int>Stack::t stack = Stack::create();	// {0}

// Raised when we reach bottom of the stack and pop is requested.
exception void Empty_stack;

// Helper function to save us typing.
void push(int x)
{
	Stack::push(stack, x);
}

// Similar ...
int pop()
{
	try {
		return Stack::pop(stack);
	} with {
	// ... but when we catch Invalid_argument() from Stack::pop
	// (indicating empty stack) ...
	case Invalid_argument(_):
		// ... we raise Empty_stack so it is cought in main().
		raise Empty_stack;
	}
}

// Given a command entered by the user perform operations on stack.
void perform_operation(string cmd)
{
	int v1, v2;
	
	switch cmd {
	case "+":	push(pop() + pop());
	case "*":	push(pop() * pop());
	case "-":
		// push(pop() - pop()) would work the other way around
		v1 = pop();
		v2 = pop();
		push(v2 - v1);
	case "/":
		// likewise
		v1 = pop();
		v2 = pop();
		push(v2 / v1);
	case "p":
		puts(itoa(Stack::top(stack)) + "\n");
	case "q":
		// Here we fake end of file when user says good bye.
		raise End_of_file;
	// If it's not any of the above ...
	case _:
		try {
			// ... it has to be a number.
			push(atoi(cmd));
		} with {
		case Invalid_argument(_):
			// But if it's not:
			puts("Invalid operation: " + cmd + "\n");
		}
	}
}

void main()
{
	for (;;)
		try {
			perform_operation(read_string());
		} with {
		case Empty_stack:
			// When we get Empty_stack, print error
			// message and continue the loop
			puts("The stack is empty.\n");
		case End_of_file:
			// But when we get End_of_file, say bye ...
			puts("Bye.\n");
			// ... and leave the program.
			Sys::exit(0);
		}
}

section init
{
	main();
}

/**
{0} There is just one thing to explain in this example: [<int>Stack::t].
The rest should be clear by now.  Gont standard libary does not include
integer stack data type.  Nor string stack or stack of stacks of ints.
Instead it provides [<'a>Stack::t], which is stack of anything.  ['a]
there is read alpha (['b] is beta, ['c] is gamma, I've got no clue
what ['foo] is :-)  

Any type could be substituted in place of ['a], hence it is called "type
variable".  [<int>Stack::t] is stack of integers, [<string>Stack::t] is
stack of strings and [<<int>Stack::t>Stack::t] is stack of stacks of ints.

Side note: there is special rule to parse [<<int>Stack::t>Stack::t]
as [< <int>Stack::t>Stack::t] (bacause [<<] is parsed as a single token). 

However it is an error to store string on int stack or vice versa.  The
type variable once bound to stack cannot be overridden.  Types of functions
manipulating stacks also depends on type of values stored in the stack.
For example after [<string>Stack::t s;] expression [Stack::pop(s)] has type
[string], while after [<int>Stack::t s;] it would have [int].

Functions can be parametrized over types as well as types themself can.
This feature of Gont is called parametric polimorphism.  It allows
fully generic programming, [Stack::t] used in our RPN calculator is good
example.  It was written once, and can be reused many times.  The stack
itself need not know anything about what values are to be stored in it.
Indeeded, there is just one copy of machine code for [<int>Stack::t],
[<string>Stack::t] and [<<<foobar>Baz::qux>list>Stack::t].

The first step is to use polimorphic types and function, writing could
come later.  You can look at the standard library reference for more
information on provided generic types.  We'll also utilize few of them
in next steps of this tutorial.
*/
