/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 876 $ $LastChangedDate: 2002-09-07 15:33:53 +0200 (Sat, 07 Sep 2002) $
 *
 * Standard library for Gont, Ptrhash module interface.
 */

/** Ptrhash provides mapping between two arbitrary types.  However mapping
    source is treated as pointer, i.e. structural equality is not tested,
    and pointer of object is used as hashing input.  */
    
/** Type of hash tables from ['a] to ['b].  */
type <'a, 'b>t;

/** Create new hash, empty table.  [size_hint] is order of initial size of 
    hash table.  Hash table will grow as necessary.  */
<'a, 'b>t create(int size_hint);

/** Add element [v] with key [k] to [h].  If there is already element with
    the same key, this new element shadows it, subsequent calls to 
    [find(h, k)] will return [v].  However after [remove(h, k)], [find(h, k)]
    will return shadowed element.  */
void add(<'a, 'b>t h, 'a k, 'b v);

/** Lookup element with key [k] in [h].  Raise [Not_found] if element is not
    in hash table.  Otherwise return it.  */
'b find(<'a, 'b>t h, 'a k);

/** Remove one element with key [k] from [h].  Raise [Not_found] if element 
    is not in hash table.  */
void remove(<'a, 'b>t h, 'a k);

/** Remove all elements with key [k] from [h].  If no such elements are found,
    do nothing.  */
void remove_all(<'a, 'b>t h, 'a k);

/** Return list of all elements with key [k] from [h].  If no such 
    elements are found return [Nil].  */
<'b>list find_all(<'a, 'b>t h, 'a k);

/** Return [true] iff element with key [k] is in [h].  */
bool mem(<'a, 'b>t h, 'a k);

/** Return number of elements stored in [h].  */
int length(<'a, 'b>t h);

/** Return copy of [h].  Time of this operation is [O(length(h))].  */
<'a, 'b>t copy(<'a, 'b>t h);

/** Same as [remove_all(h, k); add(h, k, v);].  */
void replace(<'a, 'b>t, 'a, 'b);

/** Return list of keys of elements stored in [h].  Keys are returned in
    arbitrary order, however [keys(h)] and [value(h)] will maintain the same
    order if there are no insertions nor deletions on [h] between them.  */
<'a>list keys(<'a, 'b>t h);

/** Return list of values of elements stored in [h].  See Ptrhash::keys()
    for ordering notes.  */
<'b>list values(<'a, 'b>t h);

/** Execute [f] on each element of [h], passing it its key and value.  
    It is undefined what elements will be visited if [h] is modified 
    (elements are inserted or removed) during iteration.  

    Warning: this traverses hash in *random* order, i.e. the order depends
    on random things, like where malloc() is going to give you memory.
    This might come out important, when for example doing bootstrap
    comparisons in compiler.  */
void iter(<'a, 'b>t h, *(void ('a, 'b)) f);
